"""
InternalHelp Class

Description: This is an internal web browser that is used to create a simple HTML help system.
The system assumes that the homepage to the help system is located at Help/Help.html.  The
browser has a simple interface with a home button, forward and back navigation, and options
to save the page to a PDF file or print it to the printer.

@author: Don Spickler
Last Revision: 8/14/2022

"""

from PySide2.QtCore import (Qt, QMimeData, QUrl, QDir, QMarginsF, QSize)
from PySide2.QtGui import (QIcon, QColor, QBrush, QDrag, QPageSize, QPageLayout)
from PySide2.QtWidgets import (QWidget, QAbstractItemView, QMdiSubWindow, QListWidget,
                               QTreeWidget, QVBoxLayout, QListView, QAction, QMenuBar, QTextEdit,
                               QFileDialog, QDialog, QToolBar, QMainWindow)
from PySide2.QtWebEngineWidgets import QWebEngineView
from PySide2.QtPrintSupport import (QPrintDialog, QPrinter, QPrintPreviewDialog)

import sys
import os

class InternalHelp(QMainWindow):
    """
    Internal web browser used for a simple HTML help system.
    """
    def __init__(self):
        """
        Sets up the UI for the dialog, browser window and toolbar.
        """
        super().__init__()
        self.setWindowTitle("Schedule Creator Help")
        # self.setWindowIcon(QIcon(self.mainapp.resource_path('icons/ProgramIcon.png')))

        self.view = QWebEngineView(self)
        self.view.setContextMenuPolicy(Qt.PreventContextMenu)
        self.navHome()

        self.setMinimumWidth(700)
        self.setMinimumHeight(400)

        toolbar = self.createToolBar()
        self.addToolBar(toolbar)
        self.setCentralWidget(self.view)
        self.show()

    def resource_path(self, relative_path):
        """ Adjoins a relative path for icons and help system. """
        if hasattr(sys, '_MEIPASS'):
            return os.path.join(sys._MEIPASS, relative_path)
        return os.path.join(os.path.abspath("."), relative_path)

    def createToolBar(self):
        """ Set up toolbar """
        tool_bar = QToolBar(self)
        tool_bar.setIconSize(QSize(16, 16))

        # Create actions.
        self.home_act = QAction(QIcon(self.resource_path('icons/Room.png')), "Home", self)
        self.home_act.triggered.connect(self.navHome)

        self.previous_act = QAction(QIcon(self.resource_path('icons/Previous.png')), "Back", self)
        self.previous_act.triggered.connect(self.view.back)

        self.next_act = QAction(QIcon(self.resource_path('icons/Next.png')), "Forward", self)
        self.next_act.triggered.connect(self.view.forward)

        self.savepdf_act = QAction(QIcon(self.resource_path('icons/FileSave.png')), "Save to PDF...", self)
        self.savepdf_act.triggered.connect(self.saveFile)

        self.print_act = QAction(QIcon(self.resource_path('icons/print.png')), "Print...", self)
        self.print_act.triggered.connect(self.printFile)

        # Add actions to toolbar
        tool_bar.addAction(self.home_act)
        tool_bar.addSeparator()
        tool_bar.addAction(self.previous_act)
        tool_bar.addAction(self.next_act)
        tool_bar.addSeparator()
        tool_bar.addAction(self.savepdf_act)
        tool_bar.addAction(self.print_act)

        return tool_bar

    def navHome(self):
        """ Set browser to home file. """
        helpurl = QUrl.fromLocalFile(self.resource_path("Help/Help.html"))
        self.view.load(helpurl)

    def saveFile(self):
        """ Save the current HTML view to a PDF file. """
        dialog = QFileDialog(self)
        dialog.setFilter(dialog.filter() | QDir.Hidden)
        dialog.setDefaultSuffix('pdf')
        dialog.setAcceptMode(QFileDialog.AcceptSave)
        dialog.setNameFilters(['PDF Files (*.pdf)'])
        dialog.setWindowTitle('Export to PDF')

        if dialog.exec() == QDialog.Accepted:
            filelist = dialog.selectedFiles()
            if len(filelist) > 0:
                file_name = filelist[0]
                pl = QPageLayout(QPageSize(QPageSize.Letter), QPageLayout.Portrait,
                                 QMarginsF(36, 36, 36, 36))
                self.view.page().printToPdf(file_name, pl)

    def printFile(self):
        """ Print the current HTML view. """
        printer = QPrinter()
        dialog = QPrintDialog(printer, self)
        printer.setDocName("Help")
        printer.setResolution(300)
        pl = QPageLayout(QPageSize(QPageSize.Letter), QPageLayout.Portrait,
                         QMarginsF(36, 36, 36, 36))
        printer.setPageLayout(pl)
        if dialog.exec() == QDialog.Accepted:
            self.view.page().print(printer, self.print_completed)

    def print_completed(self, success):
        """ Dummy completion function used for the QWebEngineView print command. """
        pass  # Nothing needs to be done.
